@@ -7,20 +7,24 @@
 # LICENSE file in the root directory of this source tree.
 
 import abc
-import logging
+import functools
+import json
 import socket
 import time
 from contextlib import closing
 from enum import Enum
-from typing import Any, Callable, Dict, Tuple
+from typing import Any, Callable, Dict, List, Tuple
 
 import torchelastic.rendezvous as rdzv
+import torchelastic.utils.store as store_util
 from torchelastic.metrics import prof, put_metric
+from torchelastic.utils.logging import get_logger
 
 
+_TERMINAL_STATE_SYNC_ID = "torchelastic/agent/terminal_state"
+
 DEFAULT_ROLE = "default"
-
-log = logging.getLogger(__name__)
+log = get_logger()
 
 
 class WorkerSpec:
@@ -49,8 +53,8 @@ class WorkerSpec:
         fn: Callable,
         args: Tuple,
         rdzv_handler: rdzv.RendezvousHandler,
-        max_restarts: int = 100,
-        monitor_interval: float = 5.0,
+        max_restarts: int = 3,
+        monitor_interval: float = 30.0,
         master_port=None,
     ):
         r"""
@@ -68,7 +72,7 @@ class WorkerSpec:
         """
 
         assert local_world_size > 0
-        assert max_restarts > 0
+        assert max_restarts >= 0
         assert monitor_interval > 0
 
         # Note: role is not used for data parallel, every worker has the same role
@@ -99,12 +103,28 @@ class Worker:
         id (Any): uniquely identifies a worker (interpreted by the agent)
         local_rank (int): local rank of the worker
         global_rank (int): global rank of the worker
+        role_rank (int): rank of the worker across all workers that have the same role
         world_size (int): number of workers (globally)
+        role_world_size (int): number of workers that have the same role
     """
 
-    __slots__ = ["id", "local_rank", "global_rank", "world_size"]
+    __slots__ = [
+        "id",
+        "local_rank",
+        "global_rank",
+        "role_rank",
+        "world_size",
+        "role_world_size",
+    ]
 
-    def __init__(self, local_rank: int):
+    def __init__(
+        self,
+        local_rank: int,
+        global_rank: int = -1,
+        role_rank: int = -1,
+        world_size: int = -1,
+        role_world_size: int = -1,
+    ):
         # unique identifier for this worker
         self.id: Any = None
 
@@ -112,16 +132,33 @@ class Worker:
         # by the same ``agent`` instance.
         self.local_rank: int = local_rank
 
+        #  rank of the worker among all the workers across all roles
+        #  across all ``agent`` instances.
+        #  Global rank is not stable between re-rendezvous.
+        self.global_rank: int = global_rank
+
         #  rank of the worker among all the workers with the same role
         #  across all ``agent`` instances.
         #  Global rank is not stable between re-rendezvous.
-        # pyre-fixme[8]: Attribute has type `int`; used as `None`.
-        self.global_rank: int = None
+        self.role_rank: int = role_rank
 
         # total number of workers (globally). Due to elasticity
         # the world size may change between re-rendezvous.
-        # pyre-fixme[8]: Attribute has type `int`; used as `None`.
-        self.world_size: int = None
+        self.world_size: int = world_size
+
+        # total number of workers that share the same role. Due to elasticity
+        # the role world size may change between re-rendezvous.
+        self.role_world_size: int = role_world_size
+
+    def __str__(self):
+        return (
+            f"local_rank={self.local_rank},global_rank={self.global_rank}"
+            f",role_rank={self.role_rank},world_size={self.world_size}"
+            f",role_world_size={self.role_world_size}"
+        )
+
+    def __repr__(self):
+        return str(self)
 
 
 class WorkerState(Enum):
@@ -198,6 +235,64 @@ class WorkerGroup:
         self.state = WorkerState.INIT
 
 
+class _RoleInstanceInfo:
+    """
+    The class is used by the agent to exchange the information with other agents.
+    The information is used to determine the rank of the workers that agent
+    manages in heterogeneous environments, where different agents can have
+    different number of workers.
+    """
+
+    __slots__ = ["role", "rank", "local_world_size"]
+
+    def __init__(self, role: str, rank: int, local_world_size: int):
+        r"""
+
+        Arguments:
+            role (str): user-defined role for the workers with this spec
+            rank (int): the rank of the agent
+            local_world_size (int): number of local workers to run
+        """
+
+        self.role = role
+        self.rank = rank
+        self.local_world_size = local_world_size
+
+    def serialize(self) -> bytes:
+        dict_data = {
+            "role": self.role,
+            "rank": self.rank,
+            "local_world_size": self.local_world_size,
+        }
+        return json.dumps(dict_data).encode(encoding="UTF-8")
+
+    @staticmethod
+    def deserialize(data: bytes):
+        dict_data = json.loads(data.decode(encoding="UTF-8"))
+        return _RoleInstanceInfo(
+            dict_data["role"], dict_data["rank"], dict_data["local_world_size"]
+        )
+
+    @staticmethod
+    def compare(obj1, obj2) -> int:
+        if obj1.role == obj2.role:
+            return obj1.rank - obj2.rank
+        elif obj1.role > obj2.role:
+            return 1
+        else:
+            return -1
+
+    @staticmethod
+    def find_role_boundaries(roles_infos: List, role: str) -> Tuple[int, int]:
+        start_idx, end_idx = -1, -1
+        for idx, role_info in enumerate(roles_infos):
+            if role_info.role == role:
+                if start_idx == -1:
+                    start_idx = idx
+                end_idx = idx
+        return (start_idx, end_idx)
+
+
 class MonitorResult:
     """
     Returned by the agent's ``_monitor_workers`` API. A holder object
@@ -354,9 +449,11 @@ class SimpleElasticAgent(ElasticAgent):
     for a single ``WorkerSpec`` (e.g. one particular type of worker role).
     """
 
-    def __init__(self, spec: WorkerSpec):
+    def __init__(self, spec: WorkerSpec, exit_barrier_timeout: float = 300):
         self._worker_group = WorkerGroup(spec)
         self._remaining_restarts = self._worker_group.spec.max_restarts
+        self._store = None
+        self._exit_barrier_timeout = exit_barrier_timeout
 
     # pyre-fixme[14]: `get_worker_group` overrides method defined in `ElasticAgent`
     #  inconsistently.
@@ -418,39 +515,126 @@ class SimpleElasticAgent(ElasticAgent):
         """
 
         spec = worker_group.spec
-        stride = spec.local_world_size
 
         store, group_rank, group_world_size = spec.rdzv_handler.next_rendezvous()
-        world_size = group_world_size * spec.local_world_size
+        self._store = store
 
+        workers = self._assign_worker_ranks(store, group_rank, group_world_size, spec)
+        worker_group.workers = workers
         worker_group.store = store
         worker_group.group_rank = group_rank
         worker_group.group_world_size = group_world_size
 
         if group_rank == 0:
             self._set_master_addr_port(store, spec.master_port)
-
-        assigned_global_ranks = []
-        for worker in worker_group.workers:
-            global_rank = (group_rank * stride) + worker.local_rank
-            worker.global_rank = global_rank
-            worker.world_size = world_size
-            assigned_global_ranks.append(global_rank)
-
         master_addr, master_port = self._get_master_addr_port(store)
         restart_count = spec.max_restarts - self._remaining_restarts
+        workers_info = {
+            "local_ranks": [worker.local_rank for worker in workers],
+            "global_ranks": [worker.global_rank for worker in workers],
+            "role_ranks": [worker.role_rank for worker in workers],
+            "world_size": workers[0].world_size,
+            "role_world_size": workers[0].role_world_size,
+        }
         log.info(
             f"[{spec.role}] Rendezvous complete for workers.\n"
             f"Result:\n"
             f"\trestart_count={restart_count}\n"
             f"\tgroup_rank={group_rank}\n"
             f"\tgroup_world_size={group_world_size}\n"
-            f"\trank stride={stride}\n"
-            f"\tassigned global_ranks={assigned_global_ranks}\n"
             f"\tmaster_addr={master_addr}\n"
             f"\tmaster_port={master_port}\n"
+            f"\tworkers={workers_info}\n"
         )
 
+    def _get_ranks(
+        self,
+        role_infos: List[_RoleInstanceInfo],
+        role_idx: int,
+        start_idx: int = 0,
+        end_idx: int = -1,
+    ) -> Tuple[int, List[int]]:
+        if end_idx == -1:
+            end_idx = len(role_infos)
+        prefix_sum = 0
+        total_sum = 0
+        for idx in range(start_idx, end_idx):
+            if role_idx > idx:
+                prefix_sum += role_infos[idx].local_world_size
+            total_sum += role_infos[idx].local_world_size
+        return (
+            total_sum,
+            list(range(prefix_sum, prefix_sum + role_infos[role_idx].local_world_size)),
+        )
+
+    @prof
+    def _assign_worker_ranks(
+        self, store, group_rank: int, group_world_size: int, spec: WorkerSpec
+    ) -> List[Worker]:
+        """
+        Determines proper ranks for worker processes. The rank assignment
+        is done according to the following algorithm:
+
+        1. Each agent writes its configuration(group_rank, group_world_size
+           , num_workers) to the common store.
+        2. Each agent retrieves configuration for all agents
+           and performs two level sort using role and rank.
+        3. Determine the global rank: the global rank of the workers for the current
+           agent is the offset of the infos array up to group_rank of the agent.
+           The offset is computed as a sum of local_world_size of all agents that
+           have rank less than the group_rank. The workers would have the ranks:
+           [offset, offset+local_world_size)
+        4. Determine the role rank: The role rank is determined using the algorithms
+           in the point 3 with the exception that the offset is done from the first
+           agent that has the same role as current one and has the minimum group rank.
+        """
+
+        role_infos = self._share_and_gather(store, group_rank, group_world_size, spec)
+        my_role_info = role_infos[group_rank]
+        worker_world_size, worker_global_ranks = self._get_ranks(role_infos, group_rank)
+        role_infos = sorted(
+            role_infos, key=functools.cmp_to_key(_RoleInstanceInfo.compare)
+        )
+        role_start_idx, role_end_idx = _RoleInstanceInfo.find_role_boundaries(
+            role_infos, my_role_info.role
+        )
+        role_pos = next(
+            idx
+            for idx, role_info in enumerate(role_infos)
+            if _RoleInstanceInfo.compare(role_info, my_role_info) == 0
+        )
+        role_world_size, role_ranks = self._get_ranks(
+            role_infos, role_pos, role_start_idx, role_end_idx + 1
+        )
+        workers = []
+        for ind in range(spec.local_world_size):
+            worker = Worker(
+                local_rank=ind,
+                global_rank=worker_global_ranks[ind],
+                role_rank=role_ranks[ind],
+                world_size=worker_world_size,
+                role_world_size=role_world_size,
+            )
+            workers.append(worker)
+        return workers
+
+    def _share_and_gather(
+        self, store, group_rank: int, group_world_size: int, spec: WorkerSpec
+    ) -> List:
+        agent_role_info = _RoleInstanceInfo(
+            spec.role, group_rank, spec.local_world_size
+        )
+        key_prefix = "torchelastic/role_info"
+        agent_config_enc = agent_role_info.serialize()
+        role_infos_bytes = store_util.synchronize(
+            store, agent_config_enc, group_rank, group_world_size, key_prefix
+        )
+        role_infos = [
+            _RoleInstanceInfo.deserialize(role_info_bytes)
+            for role_info_bytes in role_infos_bytes
+        ]
+        return role_infos
+
     @prof
     def _initialize_workers(self, worker_group: WorkerGroup) -> None:
         r"""
@@ -493,7 +677,58 @@ class SimpleElasticAgent(ElasticAgent):
         worker_group.state = WorkerState.STOPPED
         self._initialize_workers(worker_group)
 
+    @prof
     def run(self, role: str = DEFAULT_ROLE) -> Dict[int, Any]:
+        is_failed = False
+        try:
+            return self._invoke_run(role)
+        except Exception as e:
+            is_failed = True
+            raise
+        finally:
+            self._record_metrics(is_failed)
+
+    def _record_metrics(self, is_failed: bool = False):
+        self._record_flakiness_metric(is_failed)
+        spec = self._worker_group.spec
+        restarts_happened = self._remaining_restarts != spec.max_restarts
+        put_metric(f"workers.{spec.role}.run_total", 1)
+        self._record_metric_with_condition(
+            "run_success_with_retries", not is_failed and restarts_happened
+        )
+        self._record_metric_with_condition(
+            "run_success_no_retries", not is_failed and not restarts_happened
+        )
+        self._record_metric_with_condition(
+            "run_failed_with_retries", is_failed and restarts_happened
+        )
+        self._record_metric_with_condition(
+            "run_failed_no_retries", is_failed and not restarts_happened
+        )
+
+    def _record_metric_with_condition(self, metric_name, condition):
+        spec = self._worker_group.spec
+        if condition:
+            put_metric(f"workers.{spec.role}.{metric_name}", 1)
+        else:
+            put_metric(f"workers.{spec.role}.{metric_name}", 0)
+
+    def _record_flakiness_metric(self, is_failed: bool = False):
+        if is_failed:
+            if not isinstance(is_failed, WorkerGroupFailureException):
+                # Only user code can contribute into flakiness score
+                return
+            flakiness = 100.0
+        else:
+            spec = self._worker_group.spec
+            flakiness = 100.0 - 100.0 * (self._remaining_restarts + 1) / (
+                spec.max_restarts + 1
+            )
+        spec = self._worker_group.spec
+
+        put_metric(f"workers.{spec.role}.flakiness", int(flakiness))
+
+    def _invoke_run(self, role: str = DEFAULT_ROLE) -> Dict[int, Any]:
         # NOTE: currently only works for a single role
 
         spec = self._worker_group.spec
@@ -516,7 +751,11 @@ class SimpleElasticAgent(ElasticAgent):
             put_metric(f"workers.{role}.{state.name.lower()}", 1)
 
             if state == WorkerState.SUCCEEDED:
-                log.info(f"[{role}] All workers successfully finished.")
+                log.info(
+                    f"[{role}] worker group successfully finished."
+                    f" Waiting {self._exit_barrier_timeout} seconds for other agents to finish."
+                )
+                self._exit_barrier()
                 return monitor_result.ret_vals
             elif state in {WorkerState.UNHEALTHY, WorkerState.FAILED}:
                 if self._remaining_restarts > 0:
@@ -530,6 +769,7 @@ class SimpleElasticAgent(ElasticAgent):
                 else:
                     self._stop_workers(self._worker_group)
                     self._worker_group.state = WorkerState.FAILED
+                    self._exit_barrier()
                     raise WorkerGroupFailureException(
                         f"[{role}] exceeded max_restarts={spec.max_restarts}",
                         monitor_result.exceptions,
@@ -547,3 +787,31 @@ class SimpleElasticAgent(ElasticAgent):
                     self._restart_workers(self._worker_group)
             else:
                 raise Exception(f"[{role}] Worker group in {state.name} state")
+
+    def _exit_barrier(self):
+        """
+        Wait for ``exit_barrier_timeout`` seconds for all agents to finish
+        executing their local workers (either successfully or not). This
+        acts as a safety guard against user scripts that terminate at different
+        times. This barrier keeps the agent process alive until all workers finish.
+        """
+        log.info(
+            f"Local worker group finished ({self._worker_group.state}). "
+            f"Waiting {self._exit_barrier_timeout} seconds for other agents to finish"
+        )
+        start = time.time()
+        try:
+            store_util.barrier(
+                self._store,
+                self._worker_group.group_rank,
+                self._worker_group.group_world_size,
+                key_prefix=_TERMINAL_STATE_SYNC_ID,
+                barrier_timeout=self._exit_barrier_timeout,
+            )
+            log.info(
+                f"Done waiting for other agents. Elapsed: {time.time() - start} seconds"
+            )
+        except Exception:
+            log.exception(
+                f"Error waiting on exit barrier. Elapsed: {time.time() - start} seconds"
+            )
