@@ -21,6 +21,7 @@ with the following additional functionalities:
 1. Single-node multi-worker (with sidecar etcd server)
 
 ::
+
     >>> python -m torchelastic.distributed.launch
         --standalone
         --nnodes=1
@@ -111,20 +112,28 @@ script:
 3. ``GROUP_RANK`` - rank of the worker group. A number between 0 - ``max_nnodes``.
         When running a single worker group per node, this is the rank of the node.
 
-4. ``LOCAL_WORLD_SIZE`` - local world size (e.g. number of workers running locally).
+4. ``ROLE_RANK`` -  the rank of the worker across all the workers tha have the same
+        role. The role of the worker is specified in the ``WorkerSpec``.
+
+5. ``LOCAL_WORLD_SIZE`` - local world size (e.g. number of workers running locally).
        Equal to ``--nproc_per_node`` specified on ``torchelastic.distributed.launch``.
 
-5. ``WORLD_SIZE`` - world size (total number of workers in the job).
+6. ``WORLD_SIZE`` - world size (total number of workers in the job).
 
-6. ``MASTER_ADDR`` - fqdn of the host that is running worker with rank 0.
+7. ``ROLE_WORLD_SIZE`` - the total number of workers that was launched with the same
+        role specified in ``WorkerSpec``.
+
+8. ``MASTER_ADDR`` - fqdn of the host that is running worker with rank 0.
    Used to initialize torch distributed backend.
 
-7. ``MASTER_PORT`` - port on the ``MASTER_ADDR`` that can be used to
+9. ``MASTER_PORT`` - port on the ``MASTER_ADDR`` that can be used to
    host the tcp ``c10d`` store.
 
-8. ``TORCHELASTIC_RESTART_COUNT`` - number of worker group restarts so far.
+10. ``TORCHELASTIC_RESTART_COUNT`` - number of worker group restarts so far.
 
-9. ``TORCHELASTIC_MAX_RESTARTS`` - configured max number of restarts.
+11. ``TORCHELASTIC_MAX_RESTARTS`` - configured max number of restarts.
+
+12. ``TORCHELASTIC_RUN_ID`` - equal to rdzv run_id (e.g. unique job id).
 
 **Deployment:**
 
@@ -217,15 +226,17 @@ import uuid
 from argparse import REMAINDER, ArgumentParser
 
 import torch
-import torchelastic.rendezvous.etcd_rendezvous  # noqa: F401
-import torchelastic.rendezvous.parameters as parameters
+import torchelastic.rendezvous.registry as rdzv_registry
 from torchelastic import metrics
-from torchelastic.agent.server.api import WorkerSpec
+from torchelastic.agent.server.api import WorkerGroupFailureException, WorkerSpec
 from torchelastic.agent.server.local_elastic_agent import LocalElasticAgent
+from torchelastic.distributed.argparse_util import check_env, env
+from torchelastic.rendezvous import RendezvousParameters
 from torchelastic.rendezvous.etcd_server import EtcdServer
+from torchelastic.utils.logging import get_logger
 
 
-log = logging.getLogger(__name__)
+log = get_logger()
 
 
 def parse_args(args):
@@ -239,12 +250,14 @@ def parse_args(args):
     # worker/node size related arguments
     parser.add_argument(
         "--nnodes",
+        action=env,
         type=str,
         default="1:1",
         help="number of nodes or MIN_NODES:MAX_NODES",
     )
     parser.add_argument(
         "--nproc_per_node",
+        action=env,
         type=str,
         default="auto",
         help="number of workers per node, supported values: [auto, cpu, gpu, int]",
@@ -252,27 +265,32 @@ def parse_args(args):
 
     # rendezvous related arguments
     parser.add_argument(
-        "--rdzv_backend", type=str, default="etcd", help="rendezvous backend"
+        "--rdzv_backend",
+        action=env,
+        type=str,
+        default="etcd",
+        help="rendezvous backend",
     )
     parser.add_argument(
         "--rdzv_endpoint",
+        action=env,
         type=str,
         default="",
         help="rendezvous backend server host:port",
     )
-    parser.add_argument("--rdzv_id", type=str, help="user defined group id")
+    parser.add_argument("--rdzv_id", action=env, type=str, help="user defined group id")
     parser.add_argument(
         "--rdzv_conf",
+        action=env,
         type=str,
         default="",
         help="additional rdzv configuration (conf1=v1,conf2=v2,...)",
     )
 
-    # sidecar embed rdzv backend that defults to etcd
+    # sidecar embed rdzv backend that defaults to etcd
     parser.add_argument(
         "--standalone",
-        default=False,
-        action="store_true",
+        action=check_env,
         help="starts a local, standalone rdzv backend that is represented by"
         " etcd server on a random free port"
         "using the etcd binary specified in TORCHELASTIC_ETCD_BINARY_PATH"
@@ -285,36 +303,44 @@ def parse_args(args):
     # user-code launch related arguments
     parser.add_argument(
         "--max_restarts",
+        action=env,
         type=int,
         default=3,
         help="max number of worker group restarts before failing",
     )
     parser.add_argument(
         "--monitor_interval",
+        action=env,
         type=float,
         default=5,
         help="interval (in seconds) to monitor the state of workers",
     )
     parser.add_argument(
         "--start_method",
+        action=env,
         type=str,
         default="spawn",
         choices=["spawn", "fork", "forkserver"],
         help="multiprocessing start_method to use when creating workers",
     )
+    parser.add_argument(
+        "--role",
+        action=env,
+        type=str,
+        default="default",
+        help="user-defined role for the workers",
+    )
     parser.add_argument(
         "-m",
         "--module",
-        default=False,
-        action="store_true",
+        action=check_env,
         help="Changes each process to interpret the launch script "
         "as a python module, executing with the same behavior as"
         "'python -m'.",
     )
     parser.add_argument(
         "--no_python",
-        default=False,
-        action="store_true",
+        action=check_env,
         help='Do not prepend the training script with "python" - just exec '
         "it directly. Useful when the script is not a Python script.",
     )
@@ -349,7 +375,7 @@ def parse_min_max_nnodes(nnodes: str):
 
 
 def wrapper_fn(omp_num_threads, cmd):
-    # TODO get rid of this wrapper_fn
+    # TODO get rid of this wrapper_fn when we switch over to torchelastic.multiprocessing
     # the agent uses multiprocessing.spawn to create nproc_per_node
     # instances of fn, and hence expects fn to be a callable
     # since this launcher deals with user python scripts and executables
@@ -429,7 +455,7 @@ def main(args=None):
 
     min_nodes, max_nodes = parse_min_max_nnodes(args.nnodes)
     assert 0 < min_nodes <= max_nodes
-    assert args.max_restarts > 0
+    assert args.max_restarts >= 0
 
     if args.standalone:
         etcd_server = EtcdServer()
@@ -446,16 +472,6 @@ def main(args=None):
             f"**************************************\n"
         )
 
-    rdzv_parameters = parameters.RendezvousParameters(
-        args.rdzv_backend,
-        args.rdzv_endpoint,
-        args.rdzv_id,
-        min_nodes,
-        max_nodes,
-        args.rdzv_conf,
-    )
-
-    rdzv_handler = parameters.get_rendezvous(rdzv_parameters)
     nproc_per_node = determine_local_world_size(args.nproc_per_node)
     omp_num_threads = None
     if "OMP_NUM_THREADS" not in os.environ and nproc_per_node > 1:
@@ -485,21 +501,52 @@ def main(args=None):
     cmd.append(args.training_script)
     cmd.extend(args.training_script_args)
 
-    spec = WorkerSpec(
-        role="default",
-        local_world_size=nproc_per_node,
-        fn=wrapper_fn,
-        args=(omp_num_threads, cmd),
-        rdzv_handler=rdzv_handler,
-        max_restarts=args.max_restarts,
-        monitor_interval=args.monitor_interval,
+    rdzv_parameters = RendezvousParameters(
+        backend=args.rdzv_backend,
+        endpoint=args.rdzv_endpoint,
+        run_id=args.rdzv_id,
+        min_nodes=min_nodes,
+        max_nodes=max_nodes,
+        **_parse_rdzv_conf(args.rdzv_conf),
     )
-    metrics.initialize_metrics()
-    elastic_agent = LocalElasticAgent(spec, start_method=args.start_method)
-    elastic_agent.run(spec.role)
 
-    if args.standalone:
-        etcd_server.stop()
+    rdzv_handler = rdzv_registry.get_rendezvous_handler(rdzv_parameters)
+
+    try:
+        spec = WorkerSpec(
+            role=args.role,
+            local_world_size=nproc_per_node,
+            fn=wrapper_fn,
+            args=(omp_num_threads, cmd),
+            rdzv_handler=rdzv_handler,
+            max_restarts=args.max_restarts,
+            monitor_interval=args.monitor_interval,
+        )
+        metrics.initialize_metrics()
+        elastic_agent = LocalElasticAgent(spec, start_method=args.start_method)
+        elastic_agent.run(spec.role)
+    except WorkerGroupFailureException as e:
+        # user-code error; propagate the one from the smallest rank on this node
+        excs = e.get_worker_exceptions()
+        min_rank = min(excs.keys())
+        raise excs[min_rank] from e
+    except Exception:
+        # this is an agent (platform) error
+        raise
+    finally:
+        rdzv_handler.shutdown()
+        if args.standalone:
+            etcd_server.stop()
+
+
+def _parse_rdzv_conf(conf_str: str):
+    rdzv_conf = {}
+    confs = conf_str.split(",")
+    for kvs in confs:
+        key = kvs.split("=")[0]
+        value = kvs.split("=")[1:]
+        rdzv_conf[key] = value
+    return rdzv_conf
 
 
 if __name__ == "__main__":
